﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BinHeapArray
{
    public class MinBinaryHeapArray<T> where T : class, IComparable
    {
        public const int rootIdx = 1;
        private int countNode;
        public int CountNode { get { return countNode; } }

        private T[] binaryHeapArray;

        public MinBinaryHeapArray()
        {
            binaryHeapArray = new T[8];
            countNode = 0;
        }

        public MinBinaryHeapArray(T[] bHA)
        {
            countNode = 0;
            binaryHeapArray = new T[bHA.Length + 1];
            for (int i = 0; i < bHA.Length; i++)
            {
                binaryHeapArray[i + 1] = bHA[i];
                if (bHA[i] != null) countNode++;
            }
            Build();
        }

        private void Build()
        {
            if (binaryHeapArray.Length <= rootIdx)
                return;
            for (int index = binaryHeapArray.Length / 2 - 1; index >= rootIdx; --index)
                Heapify(index);
        }


        public void Clear()
        {
            binaryHeapArray.Initialize();
            countNode = 0;
        }

        private int Left(int index)
        {
            return index << 1;
        }

        private int Right(int index)
        {
            return (index << 1) + 1;
        }

        private int Parent(int index)
        {
            return index >> 1;
        }

        private void Swap(int aIndex, int bIndex)
        {
            binaryHeapArray[0] = binaryHeapArray[aIndex];
            binaryHeapArray[aIndex] = binaryHeapArray[bIndex];
            binaryHeapArray[bIndex] = binaryHeapArray[0];
        }

        private void Heapify(int index)
        {
            int min = index;
            int leftIndex = Left(index);
            if (leftIndex <= countNode && binaryHeapArray[leftIndex].CompareTo(binaryHeapArray[min]) < 0)
                min = leftIndex;

            int rightIndex = Right(index);
            if (rightIndex <= countNode && binaryHeapArray[rightIndex].CompareTo(binaryHeapArray[min]) < 0)
                min = rightIndex;

            if (min != index)
            {
                Swap(index, min);
                Heapify(min);
            }
        }

        private T Remove(int index)
        {
            T removedItem = binaryHeapArray[index];
            binaryHeapArray[index] = null;

            Swap(index, countNode);
            countNode--;
            if (index != countNode)
            {
                if (index == 1 || binaryHeapArray[index].CompareTo(binaryHeapArray[Parent(index)]) < 0)
                    Heapify(index);
                else
                    Decrease(index); 
            }
            
            return removedItem;
        }

        private void Decrease(int index)
        {
            // Vorrausgesetzt ist: gültiger, nicht leerer Heap 
            int parentIndex = Parent(index);
            while (index > rootIdx && binaryHeapArray[index].CompareTo(binaryHeapArray[parentIndex]) < 0)
            {
                Swap(index, parentIndex);
                index = parentIndex;
                parentIndex = Parent(index);
            }
        }

        public T ExtractMin()
        {
            return Remove(rootIdx);
        }

        public T GetMin()
        {
            return binaryHeapArray[rootIdx];
        }

        public void Add(T newNode)
        {
            if (countNode == binaryHeapArray.Length - 1)
            {
                T[] bHA = new T[binaryHeapArray.Length << 1];
                binaryHeapArray.CopyTo(bHA, 0);
                binaryHeapArray = bHA;
            }
            countNode++;
            binaryHeapArray[countNode] = newNode;
            Decrease(countNode);
        }

        public void levelorder(int start)
        {
            for (int index = start; index <= countNode; index++)
			{
                Console.Write("{0},  ", binaryHeapArray[index]);
			}
        }

        public void inorder(int index)
        {
            if (index > countNode) return;
            if (index <= countNode / 2) inorder(Left(index));
            Console.Write("{0},  ", binaryHeapArray[index]);
            if (index <= countNode / 2) inorder(Right(index));
        }

        public void preorder(int index)
        {
            if (index > countNode) return;
            Console.Write("{0},  ", binaryHeapArray[index]);
            if (index <= countNode / 2) preorder(Left(index));
            if (index <= countNode / 2) preorder(Right(index));
        }

        public void postorder(int index)
        {
            if (index > countNode) return;
            if (index <= countNode / 2) postorder(Left(index));
            if (index <= countNode / 2) postorder(Right(index));
            Console.Write("{0},  ", binaryHeapArray[index]);
        }
    }
}
